# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Christos Loizou
@date: Fri May  7 16:20:47 2021
@reference: Wu, Statistical Feature Matrix for Texture Analysis
==============================================================================
"""
import numpy as np

def con_cov_dss(f, mask, Lr, Lc, Ng=256):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else
    Lr : int
        Parameters of SFM.
    Lc : int
        Parameters of SFM.
    Ng : int, optional
        Image number of gray values. The default is 256.

    Returns
    -------
    CON : numpy ndarray
    COV : numpy ndarray
    DSS : numpy ndarray
    '''
    
    N1, N2 = f.shape
    CON = np.zeros((Lr+1,2*Lc+1),np.double) # delta contrast
    COV = np.zeros((Lr+1,2*Lc+1),np.double) # deltra covariance
    DSS = np.zeros((Lr+1,2*Lc+1),np.double) # delta dissimilarity
        
    for drow in range(Lr+1):
        for dcol in range(-Lc,Lc+1):
                
            # Define coords in statistical feature matrix
            row = drow
            col = dcol + Lc
                
            # Shift image by drow, dcol := f_d
            f_d = np.ones((N1,N2),np.double) * (Ng - 1)
            mask_d = np.zeros((N1,N2),np.double)
            if dcol < 0:
                f_d[0:(N1-drow),(-dcol):(N2)] = f[(drow):(N1),0:(N2+dcol)]
                mask_d[0:(N1-drow),(-dcol):(N2)] = mask[(drow):(N1),0:(N2+dcol)]
            else:
                f_d[0:(N1-drow),0:(N2-dcol)] = f[(drow):(N1),(dcol):(N2)]
                mask_d[0:(N1-drow),0:(N2-dcol)] = mask[(drow):(N1),(dcol):(N2)]
                            
            # Find mask to exclude pixels outside mask 
            mask_common = np.multiply(mask,mask_d)
                
            n = sum(sum(mask_common))
            f_mean = sum(sum(np.multiply(f,mask_common)))/n;
                
            # Compute CON, COV, DSS
            CON[row,col] = sum(sum(np.multiply((f-f_d),mask_common)**2))/n
            temp = np.multiply((f-f_mean),(f_d-f_mean))
            COV[row,col] = sum(sum(np.multiply(mask_common,temp)))/n
            DSS[row,col] = sum(sum(abs(np.multiply((f-f_d),mask_common))))/n
        
    # 4) Set first half on first row to zero to preserve symmetry
    CON[0,0:Lc] = 0
    COV[0,0:Lc] = 0
    DSS[0,0:Lc] = 0
    
    return CON, COV, DSS

def sfm_features(f, mask, Lr=4, Lc=4):
    '''  
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    Lr : int, optional
        Parameters of SFM. The default is 4.
    Lc : int, optional
        Parameters of SFM. The default is 4.

    Returns
    -------
    features : numpy ndarray
        1)Coarseness, 2) Contrast, 3)Periodicity, 4)Roughness.
    labels : list
        Labels of features.
    '''
    
    if mask is None:
        mask = np.ones(f.shape)
        
    # 1) Labels
    labels = ["SFM_Coarseness","SFM_Contrast","SFM_Periodicity","SFM_Roughness"]
    
    # 2) Parameters
    f = np.array(f, np.double)
    mask = np.array(mask, np.double)
    Ng = 256
         
    # 3) Calculate CON, COV, DSS
    CON, COV, DSS = con_cov_dss(f, mask, Lr, Lc, Ng)
        
    # 4) CaLculate features
    features = np.zeros(4,np.double)
    r = min([Lr,Lc]);
    features[0] = 100*(r+1)*(2*r)/sum(sum(DSS[0:(r+1),(Lc-r):(1+Lc+r)]))
    features[1] = np.sqrt(sum(sum(CON[0:2,(Lc-1):(Lc+2)]))/2)
    DSScol = DSS.reshape(-1)
    DSScol[0:(Lc+1)] = 0
    DSSmean = sum(sum(DSS)) / ((Lr+1)*(2*Lc+1)-(Lc+1))
    features[2] = (DSSmean-min(DSScol[np.nonzero(DSScol)]))/DSSmean  
    lgDSSr = np.log(DSS[0,(Lc+1):(2*Lc+1)] + 1e-16)
    lgDSSc = np.log(DSS[1:(Lr+1),(Lc)] + 1e-16)
    lgLrd = np.log(np.arange(1,Lc+1))
    lgLcd = np.log(np.arange(1,Lr+1))
    Hr = (Lc*sum(np.multiply(lgDSSr,lgLrd)) - sum(lgLrd)*sum(lgDSSr)) / (Lc*sum(lgLrd**2)-sum(lgLrd)**2)
    Hc = (Lr*sum(np.multiply(lgDSSc,lgLcd)) - sum(lgLcd)*sum(lgDSSc)) / (Lr*sum(lgLcd**2)-sum(lgLcd)**2)
    features[3] =(6-Hr-Hc)/2;
        
    return features, labels


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Thu May 20 21:16:32 2021
@reference: Thibault, Texture Indexes and Gray Level Size Zone Matrix 
            Application to Cell Nuclei Classification
==============================================================================
Note that an image and a padded image with "0" give different outputs for
many features since Ns is different (even though if we remove zeros e.g.
ps = ps[ps!=0]) and as a result jvector changes outputs.
==============================================================================
"""

import numpy as np
from skimage import measure

def glszm(f, mask):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.

    Returns
    -------
    GLSZM : numpy ndarray
        GLSZ Matrix.
    '''
    Ng = 256
    levels=np.arange(0,Ng)

    Ns = f.shape[0] * f.shape[1] + 1
    GLSZM = np.zeros((Ng-1,Ns), np.double)
    
    temp = f.copy()
    for i in range(Ng-1):
        temp[f!=levels[i]] = 0
        temp[f==levels[i]] = 1
        connected_components = measure.label(temp, connectivity=1)
        connected_components = connected_components * mask
        nZone = len(np.unique(connected_components))
        for j in range(nZone):
            col = np.count_nonzero(connected_components==j)
            GLSZM[i,col] += 1
            
    return GLSZM
            
def glszm_features(f, mask):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else.

    Returns
    -------
    features : numpy ndarray
        1)Small Zone Emphasis, 2)Large Zone Emphasis,
        3)Gray Level Nonuniformity, 4)Zone Size Nonuniformit,
        5)Zone Percentage, 6)Low Gra yLeveL Zone Emphasis,
        7)High Gray Level Zone Emphasis, 8)Small Zone Low Gray 
        Level Emphasis, 9)Small Zone High Gray LeveL Emphasis, 
        10)Large Zone Lo wGray Level Emphassis, 11)Large Zone High 
        Gray Level Emphasis, 12)Gray Level Variance,
        13)Zone Size Variance, 14)Zone Size Entropy.
    labels : list
        Labels of features.
    '''
    
    if mask is None:
        mask = np.ones(f.shape)
        
    labels = ['GLSZM_SmallZoneEmphasis', 'GLSZM_LargeZoneEmphasis',
              'GLSZM_GrayLevelNonuniformity', 'GLSZM_ZoneSizeNonuniformity',
              'GLSZM_ZonePercentage', 'GLSZM_LowGrayLeveLZoneEmphasis',
              'GLSZM_HighGrayLevelZoneEmphasis', 'GLSZM_SmallZoneLowGrayLevelEmphasis',
              'GLSZM_SmallZoneHighGrayLevelEmphasis', 'GLSZM_LargeZoneLowGrayLevelEmphassis',
              'GLSZM_LargeZoneHighGrayLevelEmphasis', 'GLSZM_GrayLevelVariance',
              'GLSZM_ZoneSizeVariance','GLSZM_ZoneSizeEntropy']
    
    P = glszm(f, mask)
    # FIXME
    #idx = np.argwhere(np.all(P[..., :] == 0, axis=0))
    #P = np.delete(P, idx, axis=1)
    
    p = P / P.sum()

    Ng, Ns = p.shape
    pg = np.sum(p,1) # Gray-Level Sum [Ng x 1]
    ps = np.sum(p,0) # Zone-Size Sum  [Ns x 1]
    jvector = np.arange(1,Ns+1) 
    ivector = np.arange(1,Ng+1)     
    Nz = np.sum(P, (0,1))
    Np = np.sum(ps * jvector, 0)
    [imat,jmat] = np.meshgrid(jvector,ivector)
        
    features = np.zeros(14, np.double)
    features[0] = np.dot(ps,((1/(jvector+1e-16))**2))
    features[1]= np.dot(ps, jvector ** 2) 
    features[2] = (pg**2).sum()
    features[3] = (ps**2).sum()
    features[4] = Nz / Np
    features[5] = np.dot(pg, 1/(ivector+1e-16)**2) 
    features[6] = np.dot(pg, ivector ** 2)
    features[7] =  np.multiply(p, 
                    np.multiply(1/(jmat+1e-16)**2,1/(imat+1e-16)**2)).sum()
    features[8] = np.multiply(p, 
                    np.multiply(jmat**2,1/(imat+1e-16)**2)).sum()
    features[9] = np.multiply(p, 
                    np.multiply(1/(jmat+1e-16)**2,imat**2)).sum()
    features[10] = np.multiply(p, np.multiply(jmat**2,imat**2)).sum()
    meang = np.dot(pg,ivector)/(Ng*Ns) 
    features[11] = ((np.multiply(p, jmat) - meang) ** 2).sum() / (Ng*Ns)
    means = np.dot(ps,jvector)/(Ng*Ns)
    features[12] = ((np.multiply(p, imat) - means) ** 2).sum() / (Ng*Ns)
    features[13] = np.multiply(p, np.log2(p+1e-16)).sum()

    return features, labels


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Wed May 12 09:51:50 2021
==============================================================================
"""

import numpy as np
import matplotlib.pyplot as plt

def histogram(f, mask, bins=32): 
    ''' 
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    bins : int, optional
         Bins for histogram. The default is 32.


    Returns
    -------
    H : numpy ndarray
        Histogram of image f for 256 gray levels.
    labels : list
        Labels of features, which are the bins' number.
    '''
    
    if mask is None:
        mask = np.ones(f.shape)
        
    f  = f.astype(np.uint8)
    mask = mask.astype(np.uint8)
    level_min = 0
    level_max = 255

    f_ravel = f.ravel() 
    mask_ravel = mask.ravel() 
    roi = f_ravel[mask_ravel.astype(bool)] 
    H = np.histogram(roi, bins=bins, range=[level_min, level_max], density=True)[0]

    labels = ['Histogram_bin_'+str(b) for b in range(bins)] 
    return H, labels
        
def plot_histogram(f, mask, Ng=256, bins=32, name=''):
    if name != '':
        name = '('+name+')'
    f_ravel = f.ravel()
    mask_ravel = mask.ravel() 
    roi = f_ravel[mask_ravel.astype(bool)]
    plt.hist(roi, bins=bins, range=(0,Ng-1), density=True)     
    plt.title('Histogram: bins='+str(bins)+' Ng='+str(Ng)+' '+name)  
    plt.show()


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Mon May 10 12:42:30 2021
@reference: Acharya, AtheromaticTM Symptomatic vs. Asymptomatic Classification
            Acharya, Carotid Ultrasound Symptomatology using Atherosclerotic Plaque
            Tsiaparas, Comparison of Multiresolution Features for Texture Classification of Carotid Atherosclerosis From B-Mode Ultrasound
==============================================================================
"""

import pywt
import numpy as np
from ..utilities import _pad_image_power_2

def dwt_features(f, mask, wavelet='bior3.3', levels=3):
    ''' 
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    wavelet : str, optional
         Filter to be used. Check pywt for filter families. The default is 'bior3.3'
    levels : int, optional
        Levels of decomposition. Default is 3.

    Returns
    -------
    features : numpy ndarray
        Mean and std of each detail image. Appromimation images are ignored.
    labels : list
        Labels of features.
    '''
    
    if mask is None:
        mask = np.ones(f.shape)
        
    # Step 1: Pad Image
    f = _pad_image_power_2(f)       # pad to the next power of 2 in each dimension
    mask = _pad_image_power_2(mask) # pad to the next power of 2 in each dimension
        
    # Step 2: Get DWT Decomposition for 3 levels
    coeffs = pywt.wavedec2(f, wavelet=wavelet, level=levels)
    coeff_arr, coeff_slices = pywt.coeffs_to_array(coeffs)
    
    # Step 3: Get DWT Decomposition for 3 levels for mask
    coeffs_mask = pywt.wavedec2(mask, wavelet=wavelet, level=levels)
    coeff_arr_mask, coeff_slices_mask = pywt.coeffs_to_array(coeffs_mask)
    coeff_arr_mask[coeff_arr_mask!=0] = 1
            
    # Step 4: For each coeff array (10-1=9 sub-images), get mean and std
    labels = []
    features = np.zeros((3*levels,2),np.double)
    i = 0
    for level in range(1,levels+1):
        for name in ['da','dd','ad']:
            D_f =  coeff_arr[coeff_slices[level][name]]
            D_mask = coeff_arr_mask[coeff_slices[level][name]]
            D = D_f.flatten()[D_mask.flatten().astype(bool)] # work on elements inside mask
            features[i][0], features[i][1] = abs(D).mean(), abs(D).std()
            i += 1
            labels.append('DWT_' + str(wavelet) + '_level_' + str(level) + 
                          '_' + str(name) + '_mean')
            labels.append('DWT_' + str(wavelet) + '_level_' + str(level) + 
                          '_' + str(name) + '_std')
            
    # Step 5: Return
    return features.flatten(), labels


 # -*- coding: utf-8 -*-
"""
==============================================================================
@author: Christos Loizou
@author: Nikolaos Giakoumoglou
@date: Fri May  7 13:53:51 2021
@reference: Amadasun, Texural Features Corresponding to Textural Properties
==============================================================================
"""
import numpy as np
from scipy import signal
from ..utilities import _image_xor

def ngtdm(f, mask, d, Ng=256):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else.
    d : int, optional
        Distance for NGTDM. Default is 1.
    Ng : int, optional
        Image number of gray values. The default is 256.

    Returns
    -------
    S : numpy ndarray
    N : numpy ndarray
    R : numpy ndarray
    '''
    
    f = f.astype(np.double)
    N1, N2 = f.shape
    oneskernel = np.ones((2*d+1,2*d+1))
    kernel = oneskernel.copy()
    kernel[d,d] = 0
    W = (2*d + 1)**2 
    
    # Get complementary mask     
    mask_c = _image_xor(mask)
    
    # Find which pixels are inside mask for convolution
    conv_mask = signal.convolve2d(mask_c,oneskernel,'same')
    conv_mask = abs(np.sign(conv_mask)-1)
        
    # Calculate abs diff between actual and neighborhood
    A = signal.convolve2d(f,kernel,'same') / (W-1)
    diff = abs(f-A)
         
    # Construct NGTDM matrix
    S = np.zeros(Ng,np.double)
    N = np.zeros(Ng,np.double)
    for x in range(d,(N1-d)):
        for y in range(d,(N2-d)):
        	if conv_mask[x,y] > 0:
        		index = f[x,y].astype('i')
        		S[index] = S[index] + diff[x,y]
        		N[index] += 1
            
    R = sum(N)
    
    return S, N, R

def ngtdm_features(f, mask, d=1):
    '''  
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    d : int, optional
        Distance for NGTDM. Default is 1.

    Returns
    -------
    features : numpy ndarray
        1)Coarseness, 2)Contrast, 3)Busyness, 4)Complexity, 5)Strength.
    labels : list
        Labels of features.
    '''
    
    if mask is None:
        mask = np.ones(f.shape)
        
    # 1) Labels
    labels = ["NGTDM_Coarseness","NGTDM_Contrast","NGTDM_Busyness",
              "NGTDM_Complexity","NGTDM_Strngth"]
    
    # 2) Parameters
    f  = f.astype(np.uint8)
    mask = mask.astype(np.uint8)
    Ng = 256
    
    # 3) Calculate NGTDM
    S, N, R = ngtdm(f, mask, d, Ng)
        
    # 4) Calculate Features
    features = np.zeros(5,np.double) 
    Ni, Nj = np.meshgrid(N,N)
    Si, Sj = np.meshgrid(S,S)
    i, j = np.meshgrid(np.arange(Ng),np.arange(Ng))
    ilessjsq = ((i-j)**2).astype(np.double)   
    Ni = np.multiply(Ni,abs(np.sign(Nj)))
    Nj = np.multiply(Nj,abs(np.sign(Ni)))     
    features[0] = R*R / sum(np.multiply(N,S))
    features[1] = sum(S)*sum(sum(np.multiply(np.multiply(Ni,Nj),ilessjsq)))/R**3/Ng/(Ng-1)
    temp = np.multiply(i,Ni) - np.multiply(j,Nj)
    features[2] = sum(np.multiply(N,S)) / sum(sum(abs(temp))) / R
    temp = np.multiply(Ni,Si) + np.multiply(Nj,Sj)
    temp2 = np.multiply(abs(i-j),temp)
    temp3 = np.divide(temp2,Ni+Nj+1e-16)
    features[3] = sum(sum(temp3)) / R
    features[4] = sum(sum(np.multiply(Ni+Nj,ilessjsq))) / (sum(S)+1e-16)
        
    return features, labels


from .fos import fos
from .glcm import glcm_features
from .glds import glds_features
from .ngtdm import ngtdm_features
from .sfm import sfm_features
from .lte import lte_measures
from .fdta import fdta
from .glrlm import glrlm_features
from .fps import fps
from .shape_parameters import shape_parameters
from .hos_v2 import hos_features, plot_sinogram
from .lbp import lbp_features
from .glszm import glszm_features

__all__ = [
    'fos',
    'glcm_features',
    'glds_features',
    'ngtdm_features',
    'sfm_features',
    'lte_measures',
    'fdta',
    'glrlm_features',
    'fps',
    'shape_parameters',
    'hos_features','plot_sinogram',
    'lbp_features',
    'glszm_features']


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Thu May 13 11:01:34 2021
@reference: Tsiaparas, Comparison of Multiresolution Features for Texture Classification of Carotid Atherosclerosis From B-Mode Ultrasound
==============================================================================
"""

import pywt
import numpy as np
from ..utilities import _pad_image_power_2

def swt_features(f, mask, wavelet='bior3.3', levels=3):
    ''' 
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    wavelet : str, optional
         Filter to be used. Check pywt for filter families. The default is 'bior3.3'
    levels : int, optional
        Levels of decomposition. Default is 3.

    Returns
    -------
    features : numpy ndarray
        Mean and std of each detail image. Appromimation images are ignored.
    labels : list
        Labels of features.
    '''
    if mask is None:
        mask = np.ones(f.shape)
        
    # Step 1: Pad image so SWT can work
    f2 = _pad_image_power_2(f)       # pad to the next power of 2 in each dimension
    mask2 = _pad_image_power_2(mask) # pad to the next power of 2 in each dimension
    
    # Step 2: Get SWT Decomposition for 3 levels
    coeffs = pywt.swt2(f2, wavelet, levels)
    
    # Step 3: For each coeff array (10-1 sub-images), get mean and std
    features = np.zeros((3*levels,2),np.double)
    D_mask = mask2.copy()
    i = 0
    for level in range(0,levels):
        coeff = coeffs[level]
        cH, cV, cD = coeff[1]
        for D_f in [cH, cV, cD]:
            D = D_f.flatten()[D_mask.flatten().astype(bool)] # work on elements inside mask
            features[i][0], features[i][1] = abs(D).mean(), abs(D).std()
            i += 1
    
    # Step 4: Create Labels
    labels = []
    for level in range(0, levels):
        for s1 in ['_h','_v','_d']:
            for s2 in ['_mean', '_std']:
                labels.append('SWT_' + str(wavelet) + '_level_' + str(level+1)
                              + str(s1) + str(s2))              
        
    # Step 5: Return
    return features.flatten(), labels


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Mon Thu May 13 13:14:35 2021
@reference: Tsiaparas, Comparison of Multiresolution Features for Texture Classification of Carotid Atherosclerosis From B-Mode Ultrasound
==============================================================================
"""

import numpy as np
from skimage.filters import gabor_kernel
from scipy import signal
from ..utilities import _image_xor

def gt_features(f, mask, deg=4, freq=[0.05, 0.4]):
    ''' 
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    deg: int, optinal
        Quantized degrees. The default is 4 (0, 45, 90, 135 degrees)
    freq: list, optional
        frequency of the gabor kernel. The default is [0.05, 0.4]

    Returns
    -------
    features : numpy ndarray
        Mean and std for the resulted image: (f o gabor_filter)(x,y)
        
    labels : list
        Labels of features.
    '''    
    
    if mask is None:
        mask = np.ones(f.shape)
        
    # Step 1: Initialize kernels
    kernels = []
    labels = []
    for theta in range(deg): # e.g. th = 0, 45, 90, 135 degrees
        theta = theta / deg * np.pi
        for frequency in freq: # e.g. f = 0.05 and 0.4
            kernel = np.real(gabor_kernel(frequency, theta=theta))
            kernels.append(kernel)
            labels.append('GT_th_' + str(theta*4/np.pi) + '_freq_' + 
                          str(frequency) + '_mean')
            labels.append('GT_th_' + str(theta*4/np.pi) + '_freq_' + 
                          str(frequency) + '_std')
            
    # Step 2: Get mask where convolution should be performed
    mask_c = _image_xor(mask)
    mask_convs = []
    for k in range(len(kernels)):
        oneskernel = np.ones(kernels[k].shape)
        temp = signal.convolve2d(mask_c, oneskernel,'same')
        temp = np.abs(np.sign(temp)-1)
        mask_convs.append(temp)
    
    # Step 3: Convolve image with each kernel and get mean and std
    features = np.zeros((len(kernels), 2), np.double)
    for k in range(len(kernels)):
            D = signal.convolve2d(f, kernels[k], 'same')
            D = np.multiply(D, mask_convs[k])
            D_ravel = D.ravel()
            mask_conv_ravel = mask_convs[k].ravel()
            roi = D_ravel[mask_conv_ravel==1]
            if roi.size == 0:
                features[k, 0], features[k, 1] = 0, 0
            else:
                features[k, 0], features[k, 1] = roi.mean(), roi.std()
                
    # step 4: Return
    return features.flatten(), labels


    
    



# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Sun May 23 13:24:21 2021
==============================================================================
"""

import numpy as np
from skimage import morphology
import itertools

def multiregion_histogram(f, mask, bins=32, num_eros=3, square_size=3):
    ''' 
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    bins : int, optional
        Bins for histogram. Default is 32.
    num_eros : int, optional
        Times of erosion to be performed. Default is 3.
    square_size : int, optional
        Kernel where erosion is performed, here a squared. Default square's
        size is 3.
        
    Returns
    -------
    features : numpy ndarray
        Histogram of f and f after erosions as a vector e.g. [32 x num_eros].
    labels : list
        Labels of features.
    '''
    
    if mask is None:
        mask = np.ones(f.shape)
    f2 = f.astype(np.uint8)               
    mask2 = mask.astype(np.uint8)   
    kernel = morphology.square(square_size)
    level_min = 0
    level_max = 255
    features = np.zeros((num_eros,bins), np.double)
    labels = []
    for i in range(num_eros):
        f2 = morphology.erosion(f2, kernel)
        mask2 = morphology.erosion(mask2, kernel)
        f2_ravel = f2.ravel()
        mask2_ravel = mask2.ravel() 
        roi = f2_ravel[mask2_ravel.astype(bool)]
        features[i,:] = np.histogram(roi, bins=bins, range=[level_min, level_max], density=True)[0]
        labels.append(['Histogram_erosion_'+str(i+1)+'_bin_'+str(b) for b in range(bins)])
    
    labels = list(itertools.chain(*labels))
    return features.flatten(), labels
    


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Thu May 13 09:50:26 2021
@reference: Ojala, A Comparative Study of Texture Measures with Classification on Feature Distributions
            Ojala, Gray Scale and Roation Invariaant Texture Classification with Local Binary Patterns
==============================================================================
"""

import numpy as np
from skimage import feature

def _energy(x):
    return np.multiply(x,x).sum()

def _entropy(x):
    return -np.multiply(x, np.log(x+1e-16)).sum()

def lbp_features(f, mask, P=[8,16,24], R=[1,2,3]):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    P : list, optional
        Number of points in neighborhood. The default is [8,16,24].
    R : list, optional
        Radius/Radii. The default is [1,2,3].

    Returns
    -------
    features : numpy ndarray
        Energy and entropy of LBP image (2 x 1).
    labels : list
        Labels of features.
    '''
    
    if mask is None:
        mask = np.ones(f.shape)
        
    P = np.array(P)
    R = np.array(R)
    n = P.shape[0]
    mask_ravel = mask.ravel() 
    features = []
    labels = []
    
    for i in range(n):
        lbp = feature.local_binary_pattern(f, P[i], R[i], 'uniform')
        lbp_ravel = lbp.ravel() 
        roi = lbp_ravel[mask_ravel.astype(bool)] 
        feats = np.zeros(2, np.double)
        feats[0] = _energy(roi) / roi.sum()
        feats[1] = _entropy(roi) / roi.sum()
        features.append(feats)
        labels.append('LBP_R_'+str(R[i])+'_P_'+str(P[i])+'_energy')
        labels.append('LBP_R_'+str(R[i])+'_P_'+str(P[i])+'_entropy')
        
    features = np.array(features, np.double).ravel()
    
    return features, labels

    


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Thu May 20 18:37:25 2021
@reference: Hu, Visual Pattern Recognition by Moment Invariants
==============================================================================
"""

import cv2

def hu_moments(f):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.

    Returns
    -------
    features : numpy ndarray
        Hu's moments.
    labels : list
        Labels of features.
    '''
    
    features = cv2.HuMoments(cv2.moments(f)).flatten()
    labels = ['Hu_Moment_' + str(i) for i in range(features.shape[0])]
    return features, labels


# -*- coding: utf-8 -*-

import numpy as np
import math
import cv2

__all__ = ['_energy', '_entropy', '_next_power_of_two', '_pad_image_power_2',
           'pad_image', '_zero_runs']

def _energy(x):
    return np.multiply(x,x).sum()

def _entropy(x):
    return -np.multiply(x, np.log(x+1e-16)).sum()

def pad_image(f, pad=2, fill_val=255):
    TDLU=[1, 1, 1, 1]  #top, down, left, right pad
    out = f.copy()
    for _ in range(pad):
        out = cv2.copyMakeBorder(out, TDLU[0], TDLU[1], TDLU[2], TDLU[3],\
                                 cv2.BORDER_CONSTANT, None, fill_val)
    return out

def _next_power_of_two(n):
    math.ceil(math.log(n,2))
    return 2 ** math.ceil(math.log(n,2))

def _pad_image_power_2(f):
    N1, N2 = f.shape
    N1_deficit = _next_power_of_two(N1) - N1
    N2_deficit = _next_power_of_two(N2) - N2
    f2 = np.pad(f, ((math.floor(N1_deficit/2),N1_deficit-math.floor(N1_deficit/2)), 
                    (math.floor(N2_deficit/2),N2_deficit-math.floor(N2_deficit/2))), 
                mode='constant')
    return f2

def _image_xor(f):
    # Turn "0" to "1" and vice versa: XOR with image consisting of "1"s
    f = f.astype(np.uint8)
    mask = np.ones(f.shape, np.uint8)
    out = np.zeros(f.shape, np.uint8)
    for i in range(f.shape[0]):
        for j in range(f.shape[1]):
            out[i,j] = f[i,j] ^ mask[i,j]
    return out

def _zero_runs(img, dimension):
    out = []
    if dimension == 0:
        dim0 = img.shape[0] 
        for dim in range(dim0):
            a = img[dim, :]
            iszero = np.concatenate(([0], np.equal(a, 0).view(np.int8), [0]))
            absdiff = np.abs(np.diff(iszero))
            ranges = np.where(absdiff == 1)[0].reshape(-1, 2)
            if ranges.size == 0:
                continue
            else:
                out.append(ranges[0][1]-ranges[0][0]+1)
    elif dimension == 1:
        dim1 = img.shape[1] 
        for dim in range(dim1):
            a = img[:, dim]
            iszero = np.concatenate(([0], np.equal(a, 0).view(np.int8), [0]))
            absdiff = np.abs(np.diff(iszero))
            ranges = np.where(absdiff == 1)[0].reshape(-1, 2)
            if ranges.size == 0:
                continue
            else:
                out.append(ranges[0][1]-ranges[0][0]+1)
    else:
        print("Error! Dimension must be 0 or 1 in an image.")
    return max(out)

    #mask_c = _image_xor(mask)
    #features[0] = _zero_runs(mask_c, 1)
    #features[1] = _zero_runs(mask_c, 0)  


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Sat May  8 12:14:26 2021
==============================================================================
"""

import numpy as np

def shape_parameters(f, mask, perimeter, pixels_per_mm2=1):
    ''' 
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    mask : numpy ndarray
        Mask image N1 x N2 with 1 if pixels belongs to ROI, 0 else. Give None
        if you want to consider ROI the whole image.
    perimeter : numpy ndarray
         Image N1 x N2 with 1 if pixels belongs to perimeter of ROI, 0 else.
    pixels_per_mm2 : int, optional
        Density of image f. The default is 1.

    Returns
    -------
    features : numpy ndarray
        1,2)X, Y coordinate max length, 3)area, 4)perimeter, 5)perimeter2/area.
    labels : list
        Labels of features.
    '''

    if mask is None:
        mask = np.ones(f.shape)
        
    # 1) Labels
    labels = ["SHAPE_XcoordMax", "SHAPE_YcoordMax", "SHAPE_area",
              "SHAPE_perimeter", "SHAPE_perimeter2perArea"]
    
    # 2) Parameters
    mask = np.array(mask, np.int32)
    perimeter = np.array(perimeter, np.int32)
    N1, N2 = mask.shape
    
    # 3) Find X max coordinate
    start, end = 0, N1
    for x in range(N1):
        if (sum(mask[x,:])>0) & (start==None):
            start = x
        if (sum(mask[x,:])==0) & (start != None):
            end = x-1
            break
    Xcoord_max = end - start + 1
    
    # 4) Find Y max coordinate
    start, end = 0, N2
    for x in range(N2):
        if (sum(mask[:,x])>0) & (start==None):
            start = x
        if (sum(mask[:,x])==0) & (start != None):
            end = x-1
            break
    Ycoord_max = end - start + 1
            
    # 5) Calculate Features
    features = np.zeros(5,np.double)
    features[0] = Xcoord_max / pixels_per_mm2
    features[1] = Ycoord_max / pixels_per_mm2
    features[2] = mask.sum() / (pixels_per_mm2 ** 2)
    features[3] = perimeter.sum() / pixels_per_mm2
    features[4] = (features[3] ** 2)/features[2]
    
    return features, labels  



from setuptools import setup, find_packages
import os

try:
    long_description = open('README.md', encoding='utf-8').read()
except:
    long_description = open('README.md').read()

classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Topic :: Software Development :: Libraries',
        "Programming Language :: Python :: 3",
	    "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]

setup(
    name="pyfeats",
    version='0.0.11',
    description='Image feature extraction inside region-of-interest',
    long_description= long_description,
    long_description_content_type = 'text/markdown',
    author="Nikolaos G. Giakoumoglou",
    author_email="<ngiakoumoglou@nhotmail.com>",
    license='MIT',
    platforms = ['Any'],
    classifiers=classifiers,
    url='https://github.com/giakou4/features', 
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    install_requires=['opencv-python', 'numpy', 'scikit-image', 'matplotlib', 'scipy', 'PyWavelets', 'mahotas'],
    keywords=['python', 'image', 'region-of-interest', 'mask', 'features', 'polygon'],
    
)


import pywt
from .dwt import dwt_features
from .swt import swt_features
from .wp import wp_features
from .gt import gt_features
from .amfm import amfm_features
from .fdta import fdta

__all__ = ['fdta',
           'dwt_features',
           'swt_features', 
           'wp_features', 
           'gt_features', 
           'amfm_features']

__wavelets__ = pywt.wavelist(family=None, kind ='all')


# -*- coding: utf-8 -*-
"""
==============================================================================
@author: Nikolaos Giakoumoglou
@date: Thu May 20 11:51:54 2021
==============================================================================
"""

from skimage import feature
import matplotlib.pyplot as plt

def hog_features(f, ppc=8, cpb=3):
    '''
    Parameters
    ----------
    f : numpy ndarray
        Image of dimensions N1 x N2.
    ppc : int, optional
        Pixels per cell. The default is 8.
    cpb : int, optional
        Cells per block. The default is 3.

    Returns
    -------
    fd : numpy ndarray
        Histogram of Oriented Gradients flattened.
    labels : list
        Labels of features.
    '''
    
    fd, _ = feature.hog(f, orientations=9, pixels_per_cell=(ppc,ppc), 
                    cells_per_block=(cpb,cpb), block_norm='L2', visualize=True)
    labels = [('HOS_ppc_' + str(ppc) + '_cpb' + str(cpb) + '_' + str(i)) for i in range(fd.shape[0])]
    return fd, labels

def plot_hog(f, ppc=16, cpb=3, name=''):
    _, hog_image = feature.hog(f, orientations=9, pixels_per_cell=(ppc,ppc), 
                    cells_per_block=(cpb,cpb), block_norm='L2', visualize=True)
    if name != '':
        name = '('+name+')'
    plt.imshow(hog_image)     
    plt.title('Histogram of Oriented Gradients: Image ' + str(name))  
    plt.show()
    


from .histogram import histogram, plot_histogram
from .multiregion_histogram import multiregion_histogram
from .correlogram import correlogram, plot_correlogram

__all__ = ['histogram', 'plot_histogram',
           'multiregion_histogram',
           'correlogram', 'plot_correlogram']